@@ -30,6 +30,7 @@
 #include "dh-preferences.h"
 #include "dh-search.h"
 #include "dh-window.h"
+#include "eggfindbar.h"
 
 struct _DhWindowPriv {
 	DhBase         *base;
@@ -44,6 +45,11 @@ struct _DhWindowPriv {
 	GtkWidget      *search;
 	GtkWidget      *html_notebook;
 
+	GtkWidget      *vbox;
+	GtkWidget      *findbar;
+	
+	gint            zoom_level;
+
 	GtkUIManager   *manager;
 	GtkActionGroup *action_group;
 };
@@ -53,6 +59,29 @@ static guint tab_accel_keys[] = {
 	GDK_6, GDK_7, GDK_8, GDK_9, GDK_0
 };
 
+static const
+struct
+{
+	gchar *name;
+	float level;
+}
+zoom_levels[] =
+{
+	{ N_("50%"), 0.7071067811 },
+	{ N_("75%"), 0.8408964152 },
+	{ N_("100%"), 1.0 },
+	{ N_("125%"), 1.1892071149 },
+	{ N_("150%"), 1.4142135623 },
+	{ N_("175%"), 1.6817928304 },
+	{ N_("200%"), 2.0 },
+	{ N_("300%"), 2.8284271247 },
+	{ N_("400%"), 4.0 }
+};
+
+#define ZOOM_MINIMAL	(zoom_levels[0].level)
+#define ZOOM_MAXIMAL	(zoom_levels[8].level)
+#define ZOOM_DEFAULT	(zoom_levels[2].level)
+
 /* People have reported problems with the default values in GConf so I'm
  * adding this to make sure that the window isn't started 1x1 pixels or the
  * paned having size 0
@@ -75,6 +104,8 @@ static void       window_activate_quit            (GtkAction       *action,
 						   DhWindow        *window);
 static void       window_activate_copy            (GtkAction       *action,
 						   DhWindow        *window);
+static void       window_activate_find            (GtkAction       *action,
+						   DhWindow        *window);
 static void       window_activate_preferences     (GtkAction       *action,
 						   DhWindow        *window);
 static void       window_activate_back            (GtkAction       *action,
@@ -87,14 +118,17 @@ static void       window_activate_show_search     (GtkAction       *action,
 						   DhWindow        *window);
 static void       window_activate_about           (GtkAction       *action,
 						   DhWindow        *window);
+static void	  window_activate_zoom_default	  (GtkAction 	   *action,
+						   DhWindow	   *window);
+static void	  window_activate_zoom_in	  (GtkAction 	   *action,
+						   DhWindow	   *window);
+static void	  window_activate_zoom_out	  (GtkAction 	   *action,
+						   DhWindow	   *window);
 static void       window_save_state               (DhWindow        *window);
 static void       window_restore_state            (DhWindow        *window);
 static gboolean   window_delete_cb                (GtkWidget       *widget,
 						   GdkEventAny     *event,
 						   gpointer         user_data);
-static gboolean   window_key_press_event_cb       (GtkWidget       *widget,
-						   GdkEventKey     *event,
-						   DhWindow        *window);
 static void       window_tree_link_selected_cb    (GObject         *ignored,
 						   DhLink          *link,
 						   DhWindow        *window);
@@ -123,6 +157,18 @@ static void       window_html_tab_accel_cb        (GtkAccelGroup   *accel_group,
 						   guint            key,
 						   GdkModifierType  mod,
 						   DhWindow        *window);
+static void       window_findbar_search_changed_cb(GObject         *object,
+						   GParamSpec      *arg1,
+						   DhWindow        *window);
+static void       window_findbar_case_sensitive_changed_cb (GObject         *object,
+							    GParamSpec      *arg1,
+							    DhWindow        *window);
+static void       window_find_previous_cb         (GtkEntry        *entry,
+						   DhWindow        *window);
+static void       window_find_next_cb             (GtkEntry        *entry,
+						   DhWindow        *window);
+static void       window_findbar_close_cb         (GtkWidget       *widget,
+						   DhWindow        *window);
 static GtkWidget *window_new_tab_label            (DhWindow        *window,
 						   const gchar     *label);
 static void       window_open_new_tab             (DhWindow        *window,
@@ -140,6 +186,7 @@ static GtkWindowClass *parent_class = NULL;
 static const GtkActionEntry actions[] = {
 	{ "FileMenu", NULL, N_("_File") },
 	{ "EditMenu", NULL, N_("_Edit") },
+	{ "ViewMenu", NULL, N_("_View") },
 	{ "GoMenu",   NULL, N_("_Go") },
 	{ "HelpMenu", NULL, N_("_Help") },
 
@@ -156,6 +203,12 @@ static const GtkActionEntry actions[] = {
 	/* Edit menu */
 	{ "Copy", GTK_STOCK_COPY, NULL, "<control>C", NULL,
 	  G_CALLBACK (window_activate_copy) },
+	{ "Find", GTK_STOCK_FIND, NULL, "<control>F", NULL,
+	  G_CALLBACK (window_activate_find) },
+	{ "Find Next", GTK_STOCK_GO_FORWARD, N_("Find Next"), "<control>G", NULL,
+	  G_CALLBACK (window_find_next_cb) },
+	{ "Find Previous", GTK_STOCK_GO_BACK, N_("Find Previous"), "<shift><control>G", NULL,
+	  G_CALLBACK (window_find_previous_cb) },
 	{ "Preferences", GTK_STOCK_PREFERENCES, NULL, NULL, NULL,
 	  G_CALLBACK (window_activate_preferences) },
 
@@ -165,15 +218,23 @@ static const GtkActionEntry actions[] = {
 	{ "Forward", GTK_STOCK_GO_FORWARD, NULL, "<alt>Right", NULL,
 	  G_CALLBACK (window_activate_forward) },
 
-	{ "ShowContentsTab", NULL, N_("Browse Contents"), "<ctrl>B", NULL,
+	{ "ShowContentsTab", NULL, N_("_Contents Tab"), "<ctrl>B", NULL,
 	  G_CALLBACK (window_activate_show_contents) },
 
-	{ "ShowSearchTab", NULL, N_("Search"), "<ctrl>S", NULL,
+	{ "ShowSearchTab", NULL, N_("_Search Tab"), "<ctrl>S", NULL,
 	  G_CALLBACK (window_activate_show_search) },
+	
+	/* View menu */
+	{ "ZoomIn", GTK_STOCK_ZOOM_IN, NULL, "<ctrl>plus", NULL,
+	  G_CALLBACK (window_activate_zoom_in) },
+	{ "ZoomOut", GTK_STOCK_ZOOM_OUT, NULL, "<ctrl>minus", NULL,
+	  G_CALLBACK (window_activate_zoom_out) },
+	{ "ZoomDefault", GTK_STOCK_ZOOM_100, NULL, "<ctrl>0", NULL,
+	  G_CALLBACK (window_activate_zoom_default) },
 
 	/* About menu */
 	{ "About", GTK_STOCK_ABOUT, NULL, NULL, NULL,
-	  G_CALLBACK (window_activate_about) }
+	  G_CALLBACK (window_activate_about) },
 };
 
 GType
@@ -226,17 +287,14 @@ window_init (DhWindow *window)
 	priv = g_new0 (DhWindowPriv, 1);
         window->priv = priv;
 
-	g_signal_connect (window,
-			  "key-press-event",
-			  G_CALLBACK (window_key_press_event_cb),
-			  window);
-
 	priv->manager = gtk_ui_manager_new ();
 
 	priv->gconf_client = gconf_client_get_default ();
 
 	accel_group = gtk_ui_manager_get_accel_group (priv->manager);
 	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
+	
+	priv->zoom_level = 2;
 
 	priv->main_box = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (priv->main_box);
@@ -295,7 +353,7 @@ static void
 window_finalize (GObject *object)
 {
 	DhWindow *window = DH_WINDOW (object);
-	DhBase *base = window->priv->base;
+	DhBase   *base = window->priv->base;
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 
@@ -353,10 +411,11 @@ window_html_switch_page_cb (GtkNotebook     *notebook,
 		new_html = g_object_get_data (G_OBJECT (new_page), "html");
 
 		window_update_title (window, new_html);
+
 		return;
 
 
-		
+
 		title = dh_html_get_title (new_html);
 		gtk_window_set_title (GTK_WINDOW (window), title);
 		g_free (title);
@@ -370,6 +429,7 @@ window_html_switch_page_cb (GtkNotebook     *notebook,
 		}
 
 		window_check_history (window, new_html);
+
 	} else {
 		gtk_window_set_title (GTK_WINDOW (window), "Devhelp");
 		window_check_history (window, NULL);
@@ -459,18 +519,48 @@ window_populate (DhWindow *window)
 			  G_CALLBACK (window_search_link_selected_cb),
 			  window);
 
+	priv->vbox = gtk_vbox_new (FALSE, 0);
+	gtk_paned_add2 (GTK_PANED (priv->hpaned), priv->vbox);
+
 	/* HTML tabs notebook. */
  	priv->html_notebook = gtk_notebook_new ();
-	gtk_paned_add2 (GTK_PANED (priv->hpaned), priv->html_notebook);
+	gtk_box_pack_start (GTK_BOX (priv->vbox), priv->html_notebook, TRUE, TRUE, 0);
 
  	g_signal_connect (priv->html_notebook,
 			  "switch-page",
 			  G_CALLBACK (window_html_switch_page_cb),
 			  window);
 
+	/* Create findbar. */
+	priv->findbar = egg_find_bar_new ();
+	gtk_widget_set_no_show_all (priv->findbar, TRUE);
+	gtk_box_pack_start (GTK_BOX (priv->vbox), priv->findbar, FALSE, FALSE, 0);
+
+	g_signal_connect (priv->findbar,
+			  "notify::search-string",
+			  G_CALLBACK(window_findbar_search_changed_cb),
+			  window);
+	g_signal_connect (priv->findbar,
+			  "notify::case-sensitive",
+			  G_CALLBACK (window_findbar_case_sensitive_changed_cb),
+			  window);
+	g_signal_connect (priv->findbar,
+			  "previous",
+			  G_CALLBACK (window_find_previous_cb),
+			  window);
+	g_signal_connect (priv->findbar,
+			  "next",
+			  G_CALLBACK (window_find_next_cb),
+			  window);
+	g_signal_connect (priv->findbar,
+			  "close",
+			  G_CALLBACK (window_findbar_close_cb),
+			  window);
+
 	dh_preferences_setup_fonts ();
 
 	gtk_widget_show_all (priv->hpaned);
+
 	window_open_new_tab (window, NULL);
 }
 
@@ -537,13 +627,80 @@ window_activate_copy (GtkAction *action, DhWindow *window)
 }
 
 static void
-window_activate_preferences (GtkAction *action, DhWindow *window)
+window_activate_find (GtkAction *action,
+		      DhWindow  *window)
+{
+	DhWindowPriv *priv;
+	DhHtml       *html;
+
+	priv = window->priv;
+
+	html = window_get_active_html (window);
+
+	dh_html_search_find (
+		html, egg_find_bar_get_search_string (EGG_FIND_BAR (priv->findbar)));
+	dh_html_search_set_case_sensitive (
+		html,  egg_find_bar_get_case_sensitive (EGG_FIND_BAR (priv->findbar)));
+
+	gtk_widget_show (priv->findbar);
+	gtk_widget_grab_focus (priv->findbar);
+}
+
+static void
+window_activate_zoom_in (GtkAction *action,
+			 DhWindow *window)
+{
+	DhWindowPriv *priv;
+	DhHtml       *html;
+
+	html = window_get_active_html (window);
+	priv = window->priv;
+
+	if (zoom_levels[priv->zoom_level].level < ZOOM_MAXIMAL) {
+		priv->zoom_level++;
+		dh_html_set_zoom (html, zoom_levels[priv->zoom_level].level);
+	}
+}
+
+static void
+window_activate_zoom_out (GtkAction *action,
+			  DhWindow *window)
+{
+	DhHtml       *html;
+	DhWindowPriv *priv;
+
+	html = window_get_active_html (window);
+	priv = window->priv;
+
+	if (zoom_levels[priv->zoom_level].level > ZOOM_MINIMAL) {
+		priv->zoom_level--;
+		dh_html_set_zoom (html, zoom_levels[priv->zoom_level].level);
+	}
+}
+
+static void
+window_activate_zoom_default (GtkAction *action,
+			      DhWindow *window)
+{
+	DhHtml       *html;
+	DhWindowPriv *priv;
+
+	html = window_get_active_html (window);
+	priv = window->priv;
+
+	dh_html_set_zoom (html, ZOOM_DEFAULT);
+}
+
+static void
+window_activate_preferences (GtkAction *action,
+			     DhWindow  *window)
 {
 	dh_preferences_show_dialog (GTK_WINDOW (window));
 }
 
 static void
-window_activate_back (GtkAction *action, DhWindow *window)
+window_activate_back (GtkAction *action,
+		      DhWindow  *window)
 {
 	DhWindowPriv *priv;
 	DhHtml       *html;
@@ -756,19 +913,6 @@ window_delete_cb (GtkWidget   *widget,
 	return FALSE;
 }
 
-static gboolean
-window_key_press_event_cb (GtkWidget   *widget,
-			   GdkEventKey *event,
-			   DhWindow    *window)
-{
-	if (event->keyval == GDK_Escape) {
-		gtk_window_iconify (GTK_WINDOW (window));
-		return TRUE;
-	}
-
-	return FALSE;
-}
-
 static void
 window_tree_link_selected_cb (GObject  *ignored,
 			      DhLink   *link,
@@ -886,6 +1030,82 @@ window_html_title_changed_cb (DhHtml      *html,
 	window_update_title (window, window_get_active_html (window));
 }
 
+static void
+window_findbar_search_changed_cb (GObject    *object,
+				  GParamSpec *pspec,
+				  DhWindow   *window)
+{
+	DhWindowPriv *priv;
+	DhHtml       *html;
+
+	priv = window->priv;
+
+	html = window_get_active_html (window);
+
+	dh_html_search_find (html,
+		egg_find_bar_get_search_string (EGG_FIND_BAR (priv->findbar)));
+}
+
+static void
+window_findbar_case_sensitive_changed_cb (GObject    *object,
+					  GParamSpec *pspec,
+					  DhWindow   *window)
+{
+	DhWindowPriv *priv;
+	DhHtml       *html;
+
+	priv = window->priv;
+
+	html = window_get_active_html (window);
+
+	dh_html_search_set_case_sensitive (
+		html,
+		egg_find_bar_get_case_sensitive (EGG_FIND_BAR (priv->findbar)));
+}
+
+static void
+window_find_next_cb (GtkEntry *entry,
+		     DhWindow *window)
+{
+	DhHtml       *html;
+	DhWindowPriv *priv = window->priv;
+
+	priv = window->priv;
+
+	gtk_widget_show (priv->findbar);
+
+	html = window_get_active_html (window);
+
+	dh_html_search_find_again (html, FALSE);
+}
+
+static void
+window_find_previous_cb (GtkEntry *entry,
+			 DhWindow *window)
+{
+	DhHtml       *html;
+	DhWindowPriv *priv = window->priv;
+
+	priv = window->priv;
+
+	gtk_widget_show (priv->findbar);
+
+	html = window_get_active_html (window);
+
+	dh_html_search_find_again (html, TRUE);
+}
+
+static void
+window_findbar_close_cb (GtkWidget *widget,
+			 DhWindow  *window)
+{
+	DhWindowPriv *priv;
+
+	priv = window->priv;
+
+	gtk_widget_hide (priv->findbar);
+}
+
 static void
 window_html_open_new_tab_cb (DhHtml      *html,
 			     const gchar *location,
@@ -941,14 +1161,16 @@ window_open_new_tab (DhWindow    *window,
 	frame = gtk_frame_new (NULL);
 	gtk_widget_show (frame);
 
+
 	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
 	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);
+
 	gtk_container_add (GTK_CONTAINER (frame), view);
 
 	g_object_set_data (G_OBJECT (frame), "html", html);
 
 	label = window_new_tab_label (window, _("Empty Page"));
-	gtk_widget_show (label);
+	gtk_widget_show_all (label);
 
 	g_signal_connect (html, "title-changed",
 			  G_CALLBACK (window_html_title_changed_cb),
@@ -982,23 +1204,90 @@ window_open_new_tab (DhWindow    *window,
 	} else {
 		dh_html_clear (html);
 	}
-
 	gtk_notebook_set_current_page (GTK_NOTEBOOK (priv->html_notebook), num);
 
 }
 
+static void
+close_button_clicked_cb (GtkButton *button,
+			 DhWindow *window)
+{
+	GtkAction *action_close;
+	
+	action_close = gtk_action_group_get_action (window->priv->action_group,
+						    "Close");
+	window_activate_close (action_close, window);
+}
+
+static void
+tab_label_style_set_cb (GtkWidget *hbox,
+			GtkStyle *previous_style,
+		        gpointer user_data)
+{
+	PangoFontMetrics *metrics;
+	PangoContext     *context;
+	GtkWidget        *button;
+	gint              char_width;
+	gint              h, w;
+
+	context = gtk_widget_get_pango_context (hbox);
+	metrics = pango_context_get_metrics (context,
+					     hbox->style->font_desc,
+					     pango_context_get_language (context));
+
+	char_width = pango_font_metrics_get_approximate_digit_width (metrics);
+	pango_font_metrics_unref (metrics);
+
+	gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (hbox),
+					   GTK_ICON_SIZE_MENU, &w, &h);
+
+	gtk_widget_set_size_request (hbox, 15 * PANGO_PIXELS (char_width) + 2 * w, -1);
+
+	button = g_object_get_data (G_OBJECT (hbox), "close-button");
+	gtk_widget_set_size_request (button, w + 2, h + 2);
+}
+
 static GtkWidget*
 window_new_tab_label (DhWindow *window, const gchar *str)
 {
-	GtkWidget *label;
+	GtkWidget  *hbox, *label, *close_button, *image;
+	GtkRcStyle *rcstyle;
+	
+	hbox = gtk_hbox_new (FALSE, 4);
 
 	label = gtk_label_new (str);
 	gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
 	gtk_label_set_single_line_mode (GTK_LABEL (label), TRUE);
 	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
 	gtk_misc_set_padding (GTK_MISC (label), 0, 0);
+	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
+	
+	close_button = gtk_button_new ();
+	gtk_button_set_relief (GTK_BUTTON (close_button),
+			       GTK_RELIEF_NONE);
+	gtk_button_set_focus_on_click (GTK_BUTTON (close_button), FALSE);
+	
+	rcstyle = gtk_widget_get_modifier_style (close_button);
+	rcstyle->xthickness = rcstyle->ythickness = 0;
+	gtk_widget_modify_style (close_button, rcstyle);
+	
+	image = gtk_image_new_from_stock (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
+	/*gtk_widget_set_tooltip_text (close_button, _("Close tab"));*/
+	g_signal_connect (close_button, "clicked",
+			  G_CALLBACK (close_button_clicked_cb), window);
+	
+	gtk_container_add (GTK_CONTAINER (close_button), image);
 
-	return label;
+	gtk_box_pack_start (GTK_BOX (hbox), close_button, FALSE, FALSE, 0);
+	
+	/* Set minimal size */
+	g_signal_connect (hbox, "style-set",
+			  G_CALLBACK (tab_label_style_set_cb), NULL);
+	
+	g_object_set_data (G_OBJECT (hbox), "label", label);
+	g_object_set_data (G_OBJECT (hbox), "close-button", close_button);
+
+	return hbox;
 }
 
 static DhHtml *
@@ -1030,7 +1319,7 @@ window_update_title (DhWindow *window,
 	gchar        *full_title;
 
 	priv = window->priv;
-	
+
 	html_title = dh_html_get_title (html);
 
 	window_tab_set_title (window, html, html_title);
@@ -1047,11 +1336,11 @@ window_update_title (DhWindow *window,
 		g_free (html_title);
 		html_title = NULL;
 	}
-	
+
 	if (!book_title) {
 		book_title = "Devhelp";
 	}
-	
+
 	if (html_title) {
 		full_title = g_strdup_printf ("%s : %s", book_title, html_title);
 		g_free (html_title);
@@ -1070,6 +1359,7 @@ window_tab_set_title (DhWindow *window, DhHtml *html, const gchar *title)
 	gint          num_pages, i;
 	GtkWidget    *view;
 	GtkWidget    *page;
+	GtkWidget    *hbox;
 	GtkWidget    *label;
 
 	priv = window->priv;
@@ -1087,10 +1377,11 @@ window_tab_set_title (DhWindow *window, DhHtml *html, const gchar *title)
 
 		/* The html widget is inside a frame. */
 		if (gtk_bin_get_child (GTK_BIN (page)) == view) {
-			label = gtk_notebook_get_tab_label (
+			hbox = gtk_notebook_get_tab_label (
 				GTK_NOTEBOOK (priv->html_notebook), page);
 
-			if (label) {
+			if (hbox) {
+				label = g_object_get_data (G_OBJECT (hbox), "label");
 				gtk_label_set_text (GTK_LABEL (label), title);
 			}
 			break;
@@ -1136,3 +1427,16 @@ dh_window_search (DhWindow *window, const gchar *str)
 	dh_search_set_search_string (DH_SEARCH (priv->search), str);
 }
 
+void
+dh_window_focus_search (DhWindow *window)
+{
+	DhWindowPriv *priv;
+
+	g_return_if_fail (DH_IS_WINDOW (window));
+
+	priv = window->priv;
+
+	gtk_notebook_set_current_page (GTK_NOTEBOOK (priv->control_notebook), 1);
+
+	dh_search_grab_focus (DH_SEARCH (priv->search));
+}
